import base64

from django.conf import settings
from django.utils.crypto import constant_time_compare

from .exceptions import Unauthorized


def get_client_ip(request):
    # IP retrieved from CloudFlare
    cf_connecting_ip = request.META.get("HTTP_CF_CONNECTING_IP")

    # Header usually set by proxies
    x_forwarded_for = request.META.get("HTTP_X_FORWARDED_FOR")

    # Header set by the connecting party, usually not the actual client making
    # the request, but a web server that the request goes through.
    remote_addr = request.META.get("REMOTE_ADDR")

    # Prioritise IPs from proxies.
    final_ip = cf_connecting_ip or x_forwarded_for or remote_addr

    # If no IP address was attached to the address, return nothing.
    if final_ip is None:
        return

    # If there is a list of IPs provided, use the last one (should be
    # the most recent one). This may not work on Google Cloud.
    return final_ip.split(",")[-1].strip()


def authorize(request, configured_username, configured_password):
    """
    Match authorization header present in the request against
    configured username and password.
    """
    # Use request.META instead of request.headers to make it
    # compatible with Django versions below 2.2.
    if "HTTP_AUTHORIZATION" not in request.META:
        raise Unauthorized('"HTTP_AUTHORIZATION" is not present in the request object.')

    authentication = request.META["HTTP_AUTHORIZATION"]

    disable_consumption = getattr(
        settings,
        "BASIC_AUTH_DISABLE_CONSUMING_AUTHORIZATION_HEADER",
        False,
    )
    if not disable_consumption:
        # Delete "Authorization" header so other authentication
        # mechanisms do not try to use it.
        request.META.pop("HTTP_AUTHORIZATION")

    authentication_tuple = authentication.split(" ", 1)
    if len(authentication_tuple) != 2:
        raise Unauthorized("Invalid format of the authorization header.")
    auth_method = authentication_tuple[0]
    auth = authentication_tuple[1]
    if "basic" != auth_method.lower():
        raise Unauthorized('"Basic" is not an authorization method.')
    auth = base64.b64decode(auth.strip()).decode("utf-8")
    username, password = auth.split(":", 1)
    username_valid = constant_time_compare(username, configured_username)
    password_valid = constant_time_compare(password, configured_password)
    if username_valid & password_valid:
        return True
    raise Unauthorized("Basic authentication credentials are invalid.")



import ipaddress

from django.conf import settings
from django.core.exceptions import PermissionDenied
from django.utils.module_loading import import_string

from .exceptions import Unauthorized
from .response import HttpUnauthorizedResponse
from .utils import authorize, get_client_ip


class BasicAuthIPWhitelistMiddleware:
    def __init__(self, get_response=None):
        self.get_response = get_response

    def __call__(self, request):
        response = None
        response = self.process_request(request)
        response = response or self.get_response(request)
        return response

    def process_request(self, request):
        # If this attribute is set, skip the check.
        if getattr(request, "_skip_basic_auth_ip_whitelist_middleware_check", False):
            return self.get_response(request)

        setattr(request, "_skip_basic_auth_ip_whitelist_middleware_check", True)
        # Check if http host is whitelisted.
        if self._is_http_host_whitelisted(request):
            return
        # Check if path is whitelisted
        if self._is_path_whitelisted(request):
            return
        # Check if IP is whitelisted
        if self._is_ip_whitelisted(request):
            return
        # Fallback to basic auth if configured.
        if self._is_basic_auth_configured():
            return self._basic_auth_response(request)
        # Otherwise just deny the access to the website
        raise PermissionDenied

    @property
    def basic_auth_login(self):
        return getattr(settings, "BASIC_AUTH_LOGIN", None)

    @property
    def basic_auth_password(self):
        return getattr(settings, "BASIC_AUTH_PASSWORD", None)

    def get_response_class(self):
        try:
            return import_string(settings.BASIC_AUTH_RESPONSE_CLASS)
        except AttributeError:
            return HttpUnauthorizedResponse

    def _basic_auth_response(self, request):
        try:
            authorize(request, self.basic_auth_login, self.basic_auth_password)
        except Unauthorized:
            return self.get_response_class()(request=request)

    def _get_client_ip(self, request):
        function_path = getattr(settings, "BASIC_AUTH_GET_CLIENT_IP_FUNCTION", None)
        func = get_client_ip
        if function_path is not None:
            func = import_string(function_path)
        return func(request)

    def _get_whitelisted_networks(self):
        networks = getattr(settings, "BASIC_AUTH_WHITELISTED_IP_NETWORKS", [])
        # If we get a list, users probably passed a list of strings in
        #  the settings, probably from the environment.
        if isinstance(networks, str):
            networks = networks.split(",")
        # Otherwise assume that the list is iterable.
        for network in networks:
            network = network.strip()
            if not network:
                continue
            yield ipaddress.ip_network(network)

    def _get_whitelisted_http_hosts(self):
        http_hosts = getattr(settings, "BASIC_AUTH_WHITELISTED_HTTP_HOSTS", [])
        # If we get a list, users probably passed a list of strings in
        #  the settings, probably from the environment.
        if isinstance(http_hosts, str):
            http_hosts = http_hosts.split(",")
        # Otherwise assume that the list is iterable.
        for http_host in http_hosts:
            http_host = http_host.strip()
            if not http_host:
                continue
            yield http_host

    def _get_whitelisted_paths(self):
        paths = getattr(settings, "BASIC_AUTH_WHITELISTED_PATHS", [])
        # If we get a list, users probably passed a list of strings in
        #  the settings, probably from the environment.
        if isinstance(paths, str):
            paths = paths.split(",")
        # Otherwise assume that the list is iterable.
        for path in paths:
            path = path.strip()
            if not path:
                continue
            yield path

    def _is_http_host_whitelisted(self, request):
        request_host = request.get_host()
        if not request_host:
            return False
        return request_host in self._get_whitelisted_http_hosts()

    def _is_path_whitelisted(self, request):
        """
        Check if request.path is whitelisted. Subpaths are included.
        """
        for path in self._get_whitelisted_paths():
            if request.path.startswith(path):
                return True
        return False

    def _is_ip_whitelisted(self, request):
        """
        Check if IP is on the whitelisted network.
        """
        ip_address = ipaddress.ip_address(self._get_client_ip(request))
        for network in self._get_whitelisted_networks():
            if ip_address in network:
                return True
        return False

    def _is_basic_auth_configured(self):
        """
        Check basic authentication username and password are
        configured.
        """
        return self.basic_auth_login and self.basic_auth_password



from django.conf import settings
from django.http import HttpResponse
from django.template.loader import render_to_string
from django.utils.cache import add_never_cache_headers

DEFAULT_AUTH_TEMPLATE = (
    "<title>Authentication Required</title><h1>Authentication required</h1>"
)


class HttpUnauthorizedResponse(HttpResponse):
    def __init__(self, content=None, request=None, *args, **kwargs):
        self._request = request
        self._content = content
        kwargs.setdefault("content_type", "text/html")
        kwargs.setdefault("status", 401)
        super().__init__(self.get_response_content(), *args, **kwargs)
        self["WWW-Authenticate"] = self.get_www_authenticate_value()
        add_never_cache_headers(self)

    def get_www_authenticate_value(self):
        value = "Basic"
        realm = getattr(settings, "BASIC_AUTH_REALM", "")
        if realm:
            realm = realm.replace('"', '\\"')
            value += ' realm="{realm}"'.format(realm=realm)
        return value

    def get_response_content(self):
        if self._content:
            return self._content
        try:
            template = settings.BASIC_AUTH_RESPONSE_TEMPLATE
        except AttributeError:
            return DEFAULT_AUTH_TEMPLATE
        return render_to_string(template, {}, request=self._request)
